/* Internationalization implementation.  Includes definitions of English
 * string arrays, and the i18n pointer. */

#include <linux/slab.h>               /* For kmalloc. */
#include <linux/ctype.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/input.h>
#include "speakup.h"
#include "spk_priv.h"

char *dflt_announce_messages[] = {
	[MSG_BLANK] = "blank",
	[MSG_IAM_ALIVE] = "I'm aLive!",
	[MSG_YOU_KILLED_SPEAKUP] = "You killed speakup!",
	[MSG_HEY_THATS_BETTER] = "hey. That's better!",
	[MSG_YOU_TURNED_ME_OFF] = "You turned me off!",
	[MSG_PARKED] = "parked!",
	[MSG_UNPARKED] = "unparked!",
	[MSG_MARK] = "mark",
	[MSG_MARK_CLEARED] = "mark, cleared",
	[MSG_OFF] = "off",
	[MSG_ON] = "on",
	[MSG_NO_WINDOW] = "no window",
	[MSG_EDIT_DONE] = "edit done",
	[MSG_WINDOW_ALREADY_SET] = "window already set, clear then reset",
	[MSG_END_BEFORE_START] = "error end before start",
	[MSG_WINDOW_CLEARED] = "window cleared",
	[MSG_WINDOW_SILENCED] = "window silenced",
	[MSG_WINDOW_SILENCE_DISABLED] = "window silence disabled",
	[MSG_WINDOW_LONG_HEADER] = "header too long",
	[MSG_ERROR] = "error",
	[MSG_COLUMNS] = "columns",
	[MSG_GOTO] = "go to",
	[MSG_ROUTE_FAILED] = "route cursor failed",
	[MSG_INDENT] = "indent",
	[MSG_IS_UNASSIGNED] = "is unassigned",
	[MSG_HELP_INFO] = "press space to leav help, cursor up or down to scroll, or a letter to go to commands in list",
	[MSG_DONE_HELP] = "leaving help",
	[MSG_VARS_INFO] = "press space to leav vars, cursor up or down to scroll, cursor left or right to adjust",
	[MSG_DONE_VARS] = "vars done",
	[MSG_EDGE_TOP] = "top,",
	[MSG_EDGE_BOTTOM] = "bottom,",
	[MSG_EDGE_LEFT] = "left,",
	[MSG_EDGE_RIGHT] = "right,",
	[MSG_MULTI] = "set to multi",
	[MSG_SINGLE] = "set to single",
	[MSG_NUMBER] = "number",
	[MSG_SPACE] = "space",
	[MSG_START] = "start",
	[MSG_END] = "end",
	[MSG_HEADER_START] = "header start",
	[MSG_HEADER_END] = "header end",
	[MSG_CTRL] = "control-",
	[MSG_NOT_FOUND] = "not found",
	[MSG_NO_BUFF] = "empty buffer",
	[MSG_DISJUNCTION] = "or",
};
#define NUM_ANNOUNCE ARRAY_SIZE(dflt_announce_messages)
char *announce_messages[NUM_ANNOUNCE];

char *dflt_format_messages[]= { // message ids are +256 of this array
"line %ld, col %ld, t t y %d",
"hex %02x, decimal %d",
"times %d .",
"repeated %d .",
"window is line %d",
"%s at line %d, column %d",
"edit  %s, press space when done",
"no commands for %c",
"is %s",
	"%s canceled",
};
#define NUM_FORMATS ARRAY_SIZE(dflt_format_messages)
char *format_messages[NUM_FORMATS];

char *dflt_color_messages[] = {
"black", "blue", "green",
"cyan", "red", "magenta",
"yellow", "white", "grey",
"bright", "on blinking"
};
#define NUM_COLORS ARRAY_SIZE(dflt_color_messages)
char *color_messages[NUM_COLORS];

char *dflt_control_messages[] = {
	"shift", "altgr", "control",
	"ault", "l shift", "speakup",
	"l control", "r control", "caps shift"
};
#define NUM_CONTROLS ARRAY_SIZE(dflt_control_messages)
char *control_messages[NUM_CONTROLS];

char *dflt_state_messages[] = {
	"shift", "altgr", "ctrl",
	"alt", "speakup", "double",
};
#define  NUM_STATES ARRAY_SIZE(dflt_state_messages)
char *state_messages[NUM_STATES];

char *dflt_func_messages[] = { // index is the speakup key
	[ATTRIB_BLEEP_DEC] = "attribute bleep decrement",
	[ATTRIB_BLEEP_INC] = "attribute bleep increment",
	[BLEEPS_DEC] = "bleeps decrement",
	[BLEEPS_INC] = "bleeps increment",
	[SAY_CHAR_ABOVE] = "character, say above",
	[SAY_CHAR_BELOW] = "character, say below",
	[SAY_CHAR] = "character, say current",
	[SAY_FIRST_CHAR] = "character, first",
	[SAY_CHAR_NUM] = "character, say hex and decimal",
	[SAY_LAST_CHAR] = "character, last",
	[SAY_NEXT_CHAR] = "character, say next",
	[SAY_PHONETIC_CHAR] = "character, say phonetic",
	[SAY_PREV_CHAR] = "character, say previous",
	[TOGGLE_COLUMN] = "toggle columns, reading cursor should be on headers",
	[TOGGLE_CURSORING] = "cursoring",
	[CURSOR_OFF] = "cursoring off",
	[CURSOR_ON] = "cursoring on",
	[CURSOR_SMART] = "smart cursoring",
	[SPEAKUP_PARKED] = "cursor park",
	[CUT] = "cut local",
	[CUT1] = "cut1",
	[CUT2] = "cut2",
	[CUT3] = "cut3",
	[EDIT_DELIM] = "edit delimiters",
	[EDIT_EXNUM] = "edit exnum",
	[EDIT_MOST] = "edit most",
	[EDIT_REPEAT] = "edit repeats",
	[EDIT_SOME] = "edit some",
	[FIND_PREV] = "find previous",
	[FIND_NEXT] = "find next",
	[FIND_UP] = "find up",
	[FIND_DOWN] = "find down",
	[GOTO] = "go to",
	[BOTTOM_EDGE] = "go to bottom edge",
	[LEFT_EDGE] = "go to left edge",
	[RIGHT_EDGE] = "go to right edge",
	[TOP_EDGE] = "go to top edge",
	[SPEAKUP_HELP] = "help",
	[CURSOR_HIGHLIGHT] = "highlight tracking",
	[TOGGLE_INDENT] = "toggle indent",
	[SAY_LINE] = "line, say current",
	[SAY_LINE_NC] = "line, say current, no column mode",
	[SAY_NEXT_LINE] = "line, say next",
	[SAY_NEXT_LINE_NC] = "line, say next, no column mode",
	[SAY_PREV_LINE] = "line, say previous",
	[SAY_PREV_LINE_NC] = "line, say previous, no column mode",
	[TOGGLE_MULTI] = "toggle multi single",
	[PASTE] = "paste local",
	[PASTE1] = "paste1",
	[PASTE2] = "paste2",
	[PASTE3] = "paste3",
	[PASTE_READ] = "read paste local",
	[PASTE_READ1] = "read paste1",
	[PASTE_READ2] = "read paste2",
	[PASTE_READ3] = "read paste3",
	[PITCH_DEC] = "pitch decrement",
	[PITCH_INC] = "pitch increment",
	[PUNCT_DEC] = "punctuation decrement",
	[PUNCT_INC] = "punctuation increment",
	[PUNC_LEVEL_DEC] = "punc level decrement",
	[PUNC_LEVEL_INC] = "punc level increment",
	[SPEAKUP_QUIET] = "quiet",
	[RATE_DEC] = "rate decrement",
	[RATE_INC] = "rate increment",
	[READING_PUNC_DEC] = "reading punctuation decrement",
	[READING_PUNC_INC] = "reading punctuation increment",
	[READ_ALL] = "read all",
	[READ_ALL_LINE] = "read all line mode",
	[CURSOR_WIN] = "read windo",
	[ROUTE_CURSOR] = "route cursor",
	[SAY_ATTRIBUTES] = "say attributes",
	[SAY_FROM_LEFT] = "say from left",
	[SAY_FROM_TOP] = "say from top",
	[SAY_LINE_INDENT] = "line, say with indent",
	[SAY_POSITION] = "say position",
	[SAY_SCREEN] = "say screen",
	[SAY_TO_BOTTOM] = "say to bottom",
	[SAY_TO_RIGHT] = "say to right",
	[SPK_KEY] = "speakup",
	[SPK_LOCK] = "speakup lock",
	[SPEAKUP_OFF] = "speakup off",
	[SPEECH_KILL] = "speech kill",
	[SPELL_DELAY_DEC] = "spell delay decrement",
	[SPELL_DELAY_INC] = "spell delay increment",
	[SPELL_WORD] = "spell word",
	[SPELL_PHONETIC] = "spell word phoneticly",
	[TONE_DEC] = "tone decrement",
	[TONE_INC] = "tone increment",
	[SET_VARS] = "set vars",
	[VOICE_DEC] = "voice decrement",
	[VOICE_INC] = "voice increment",
	[VOL_DEC] = "volume decrement",
	[VOL_INC] = "volume increment",
	[CLEAR_WIN] = "window, clear",
	[SET_WIN_HEADER] = "window, set header",
	[SAY_WIN] = "window, say",
	[SET_WIN] = "window, set",
	[ENABLE_WIN] = "window, silence",
	[WORD_END] = "word, go to end",
	[SAY_WORD_ABOVE] = "word, say above",
	[SAY_WORD_BELOW] = "word, say below",
	[SAY_WORD] = "word, say current",
	[SAY_NEXT_WORD] = "word, say next",
	[SAY_PREV_WORD] = "word, say previous"
};
#define NUM_FUNCS ARRAY_SIZE(dflt_func_messages)
char *func_messages[NUM_FUNCS];

char *dflt_key_messages[] = {
	[KEY_ESC] = "escape",
	[KEY_1] = "1",
	[KEY_2] = "2",
	[KEY_3] = "3",
	[KEY_4] = "4",
	[KEY_5] = "5",
	[KEY_6] = "6",
	[KEY_7] = "7",
	[KEY_8] = "8",
	[KEY_9] = "9",
	[KEY_0] = "0",
	[KEY_MINUS] = "minus",
	[KEY_EQUAL] = "equal",
	[KEY_BACKSPACE] = "back space",
	[KEY_TAB] = "tab",
	[KEY_Q] = "q",
	[KEY_W] = "w",
	[KEY_E] = "e",
	[KEY_R] = "r",
	[KEY_T] = "t",
	[KEY_Y] = "y",
	[KEY_U] = "u",
	[KEY_I] = "i",
	[KEY_O] = "o",
	[KEY_P] = "p",
	[KEY_LEFTBRACE] = "left brace",
	[KEY_RIGHTBRACE] = "right brace",
	[KEY_ENTER] = "enter",
	[KEY_LEFTCTRL] = "left control",
	[KEY_A] = "a",
	[KEY_S] = "s",
	[KEY_D] = "d",
	[KEY_F] = "f",
	[KEY_G] = "g",
	[KEY_H] = "h",
	[KEY_J] = "j",
	[KEY_K] = "k",
	[KEY_L] = "l",
	[KEY_SEMICOLON] = "semicolon",
	[KEY_APOSTROPHE] = "apostrophe",
	[KEY_GRAVE] = "accent",
	[KEY_LEFTSHIFT] = "left shift",
	[KEY_BACKSLASH] = "back slash",
	[KEY_Z] = "z",
	[KEY_X] = "x",
	[KEY_C] = "c",
	[KEY_V] = "v",
	[KEY_B] = "b",
	[KEY_N] = "n",
	[KEY_M] = "m",
	[KEY_COMMA] = "comma",
	[KEY_DOT] = "dot",
	[KEY_SLASH] = "slash",
	[KEY_RIGHTSHIFT] = "right shift",
	[KEY_KPASTERISK] = "keypad asterisk",
	[KEY_LEFTALT] = "left alt",
	[KEY_SPACE] = "space",
	[KEY_CAPSLOCK] = "caps lock",
	[KEY_F1] = "f1",
	[KEY_F2] = "f2",
	[KEY_F3] = "f3",
	[KEY_F4] = "f4",
	[KEY_F5] = "f5",
	[KEY_F6] = "f6",
	[KEY_F7] = "f7",
	[KEY_F8] = "f8",
	[KEY_F9] = "f9",
	[KEY_F10] = "f10",
	[KEY_NUMLOCK] = "num lock",
	[KEY_SCROLLLOCK] = "scroll lock",
	[KEY_KP7] = "keypad 7",
	[KEY_KP8] = "keypad 8",
	[KEY_KP9] = "keypad 9",
	[KEY_KPMINUS] = "keypad minus",
	[KEY_KP4] = "keypad 4",
	[KEY_KP5] = "keypad 5",
	[KEY_KP6] = "keypad 6",
	[KEY_KPPLUS] = "keypad plus",
	[KEY_KP1] = "keypad 1",
	[KEY_KP2] = "keypad 2",
	[KEY_KP3] = "keypad 3",
	[KEY_KP0] = "keypad 0",
	[KEY_KPDOT] = "keypad dot",
	[84] = "103rd", // not in input-event-codes.h
	[KEY_F13] = "f13",
	[KEY_102ND] = "102nd",
	[KEY_F11] = "f11",
	[KEY_F12] = "f12",
	[KEY_F14] = "f14",
	[KEY_F15] = "f15",
	[KEY_F16] = "f16",
	[KEY_F17] = "f17",
	[KEY_F18] = "f18",
	[KEY_F19] = "f19",
	[KEY_F20] = "f20",
	[KEY_KPENTER] = "keypad enter",
	[KEY_RIGHTCTRL] = "right control",
	[KEY_KPSLASH] = "keypad slash",
	[KEY_SYSRQ] = "sysrq",
	[KEY_RIGHTALT] = "right alt",
	[KEY_LINEFEED] = "line feed",
	[KEY_HOME] = "home",
	[KEY_UP] = "up",
	[KEY_PAGEUP] = "page up",
	[KEY_LEFT] = "left",
	[KEY_RIGHT] = "right",
	[KEY_END] = "end",
	[KEY_DOWN] = "down",
	[KEY_PAGEDOWN] = "page down",
	[KEY_INSERT] = "insert",
	[KEY_DELETE] = "delete",
	[KEY_MACRO] = "macro",
	[KEY_MUTE] = "mute",
	[KEY_VOLUMEDOWN] = "volume down",
	[KEY_VOLUMEUP] = "volume up",
	[KEY_POWER] = "power",
	[KEY_KPEQUAL] = "keypad equal",
	[KEY_KPPLUSMINUS] = "keypad plusminus",
	[KEY_PAUSE] = "pause",
	[KEY_F21] = "f21",
	[KEY_F22] = "f22",
	[KEY_F23] = "f23",
	[KEY_F24] = "f24",
	[KEY_KPCOMMA] = "keypad comma",
	[KEY_LEFTMETA] = "left meta",
	[KEY_RIGHTMETA] = "right meta",
	[KEY_COMPOSE] = "compose",
	[KEY_STOP] = "stop",
	[KEY_AGAIN] = "again",
	[KEY_PROPS] = "props",
	[KEY_UNDO] = "undo",
	[KEY_FRONT] = "front",
	[KEY_COPY] = "copy",
	[KEY_OPEN] = "open",
	[KEY_PASTE] = "paste",
	[KEY_FIND] = "find",
	[KEY_CUT] = "cut",
	[KEY_HELP] = "help",
	[KEY_MENU] = "menu",
	[KEY_CALC] = "calc",
	[KEY_SETUP] = "setup",
	[KEY_SLEEP] = "sleep",
	[KEY_WAKEUP] = "wakeup",
	[KEY_FILE] = "file",
	[KEY_SENDFILE] = "send file",
	[KEY_DELETEFILE] = "delete file",
	[KEY_XFER] = "transfer",
	[KEY_PROG1] = "prog1",
	[KEY_PROG2] = "prog2",
	[KEY_WWW] = "www",
	[KEY_MSDOS] = "msdos",
	[KEY_COFFEE] = "coffee",
	[KEY_DIRECTION] = "direction",
	[KEY_CYCLEWINDOWS] = "cycle windows",
	[KEY_MAIL] = "mail",
	[KEY_BOOKMARKS] = "bookmarks",
	[KEY_COMPUTER] = "computer",
	[KEY_BACK] = "back",
	[KEY_FORWARD] = "forward",
	[KEY_CLOSECD] = "close cd",
	[KEY_EJECTCD] = "eject cd",
	[KEY_EJECTCLOSECD] = "eject close cd",
	[KEY_NEXTSONG] = "next song",
	[KEY_PLAYPAUSE] = "play pause",
	[KEY_PREVIOUSSONG] = "previous song",
	[KEY_STOPCD] = "stop cd",
	[KEY_RECORD] = "record",
	[KEY_REWIND] = "rewind",
	[KEY_PHONE] = "phone",
	[KEY_ISO] = "iso",
	[KEY_CONFIG] = "config",
	[KEY_HOMEPAGE] = "home page",
	[KEY_REFRESH] = "refresh",
	[KEY_EXIT] = "exit",
	[KEY_MOVE] = "move",
	[KEY_EDIT] = "edit",
	[KEY_SCROLLUP] = "scroll up",
	[KEY_SCROLLDOWN] = "scroll down",
	[KEY_KPLEFTPAREN] = "keypad left paren",
	[KEY_KPRIGHTPAREN] = "keypad right paren"
};
#define NUM_KEYS ARRAY_SIZE(dflt_key_messages)
char *key_messages[NUM_KEYS];

static tmsg_group all_groups[] = {
	[ANNOUNCE_ID] ={
		.name = "announcements",
		.defaults = dflt_announce_messages,
		.messages =  announce_messages,
		.group_id = ANNOUNCE_ID,
		.size =  NUM_ANNOUNCE
	},
	[FORMAT_ID] = {
		.name = "formatted",
		.defaults = dflt_format_messages,
		.messages = format_messages,
		.group_id = FORMAT_ID,
		.size = NUM_FORMATS
	},
	{
		.name = "colors",
		.defaults = dflt_color_messages,
		.messages = color_messages,
		.group_id = COLOR_ID,
		.size = NUM_COLORS
	},
	[CONTROL_ID] = {
		.name = "ctl_keys",
		.defaults = dflt_control_messages,
		.messages = control_messages,
		.group_id = CONTROL_ID,
		.size = NUM_CONTROLS
	},
	[STATE_ID] = {
		.name = "states",
		.defaults = dflt_state_messages,
		.messages = state_messages,
		.group_id = STATE_ID,
		.size = NUM_STATES
},
	[FUNC_ID] = {
		.name = "function_names",
		.defaults = dflt_func_messages,
		.messages = func_messages,
		.group_id = FUNC_ID,
		.size = NUM_FUNCS
},
	[KEY_ID] = {
		.name = "key_names",
		.defaults = dflt_key_messages,
		.messages = key_messages,
		.group_id = KEY_ID,
		.size = NUM_KEYS
}
};

static const  int num_groups = ARRAY_SIZE(all_groups);

char *spk_msg_get(enum msg_index_t msg)
{
	int id = msg>>8, m = msg&255;
	pmsg_group group;
	if (id < ANNOUNCE_ID || id >= MAX_ID)
		return NULL;
	group = &all_groups[id];
	return (m >= 0 && m < group->size) ? group->messages[m] : NULL;
}

char *spk_msg_get2(enum msg_group_t group_id, int msg)
{
	pmsg_group group;
	if (group_id < ANNOUNCE_ID || group_id >= MAX_ID)
		return NULL;
	group = &all_groups[group_id];
	return (msg >= 0 && msg < group->size) ? group->messages[msg] : NULL;
}

void spk_msg_say(enum msg_index_t msg)
{
	int id = msg>>8, m = msg&255;
	pmsg_group group;
	if (id < ANNOUNCE_ID || id >= MAX_ID)
		return;
	group = &all_groups[id];
	if (m < 0 || m >= group->size)
		return;
	if (group->messages[m])
		synth_printf( "%s\n", group->messages[m]);
}

void spk_msg_say_sp(enum msg_index_t msg)
{
	int id = msg>>8, m = msg&255;
	pmsg_group group;
	if (id < ANNOUNCE_ID || id >= MAX_ID)
		return;
	group = &all_groups[id];
	if (m < 0 || m >= group->size)
		return;
	if (group->messages[m])
		synth_printf( "%s ", group->messages[m]);
}

void spk_msg_say2(enum msg_group_t group_id, int msg)
{
	pmsg_group group;
	if (group_id < ANNOUNCE_ID || group_id >= MAX_ID)
		return;
	group = &all_groups[group_id];
	if (msg < 0 || msg >= group->size)
		return;
	if (group->messages[msg])
		synth_printf( " %s ", group->messages[msg]);
}


/*
 * Function: next_specifier
 * Finds the start of the next format specifier in the argument string.
 * Return value: pointer to start of format specifier, or NULL
*/
static char *next_specifier(char *input)
{
	char *np = input;

	while (( np = strchr(np, '%'))) {
		if (np[1] != '%')
			break;
		np += 2;
	}
	return np;
}


/*
 * Function: find_specifier_end
 * Return a pointer to the end of the format specifier.
*/
static char *find_specifier_end(char *input)
{
	input++;		/* Advance over %. */
	while ( *input && strchr(" 0+-#", *input))
		input++; // flags
	while (isdigit(*input))
		input++; // width
	if (*input == '.') {
		input++;
		while (isdigit(*input))
			input++; // decimal
	}
	if ((input[0] == 'l') && (input[1] == 'd'))
		input += 2;
	else if ( *input && strchr("cdsx", *input))
		input++;
	return input;
}


/*
 * Function: fmt_validate
 * Check that two format strings contain the same number of format specifiers,
 * and that the order of specifiers is the same in both strings. 
 * Return 1 if the condition holds, 0 if it doesn't.
*/
static int fmt_validate(char *template, char *user)
{
	char *end1, *end2;
	size_t length1, length2;
	int valid = 1;
	do {
		template = next_specifier(template);
		user = next_specifier(user);
		if (!template || !user)
			break;
		end1 = find_specifier_end(template);
		end2 = find_specifier_end(user);
		length1 = end1-template;
		length2 = end2-user;
		if((length1 != length2) || memcmp(template, user, length1))
			valid = 0;
		template = end1;
		user = end2;
	} while (valid);
			return (template || user) ? 0 : 1;
	// if same number, and all compared, both pointers will be null
}

/*
 * Function: msg_set
 * Description: Add a user-supplied message to the user_messages array.
 * The message text is copied to a memory area allocated with kmalloc.
 * If the function fails, then user_messages is untouched.
 * Failure conditions:
 * -EINVAL -  Invalid format specifiers in formatted message or illegal index.
 * -ENOMEM -  Unable to allocate memory.
*/
ssize_t spk_msg_set(pmsg_group group, int msg, char *text, size_t length)
{
	char *newstr = NULL, *mp;
	unsigned long flags;
	int start = 0, len;
	if (FUNC_ID == group->group_id || KEY_ID == group->group_id)
		start = 1;
	if (msg < start || msg >= group->size)
		return -EINVAL;
	spin_lock_irqsave(&speakup_info.spinlock, flags);
	mp = group->messages[msg];
	len = strlen(mp);
	if (len == length && !strncmp(mp, text, length)) {
		spin_unlock_irqrestore(&speakup_info.spinlock, flags);
		return 1;
	}
	if ( !(newstr = kmalloc(length + 1, GFP_KERNEL))) {
		spin_unlock_irqrestore(&speakup_info.spinlock, flags);
		return -ENOMEM;
	}
	memcpy(newstr, text, length);
	newstr[length] = '\0';
	if (FORMAT_ID == group->group_id &&
		!fmt_validate(group->defaults[msg], text)) {
		kfree(newstr);
		spin_unlock_irqrestore(&speakup_info.spinlock, flags);
		return -EINVAL;
	}
	if (mp != group->defaults[msg])
		kfree(mp);
	group->messages[msg]= newstr;
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
	return 0;
}

pmsg_group spk_get_msg_group(enum msg_group_t group_id)
{
	if (group_id < ANNOUNCE_ID || group_id >= MAX_ID)
		return NULL;
	return &all_groups[group_id];
}

/*
 * Find a message group, given its name.  Return a pointer to the structure
 * if found, or NULL otherwise.
*/
pmsg_group spk_find_msg_group(const char *group_name)
{
	pmsg_group group = NULL;
	enum msg_group_t i;

	for (i = 0; i < MAX_ID; i++) {
		if (!strcmp(all_groups[i].name, group_name)) {
			group = &all_groups[i];
			break;
		}
	}
	return group;
}

void spk_reset_msg_group(pmsg_group group)
{
	unsigned long flags;
	int i;

	spin_lock_irqsave(&speakup_info.spinlock, flags);
	for (i = 0; i < group->size; i++) {
			if (group->messages[i] && group->messages[i] != group->defaults[i])
			kfree(group->messages[i]);
		group->messages[i] = group->defaults[i];
	}
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}

/* Called at initialization time, to establish default messages. */
void spk_initialize_msgs(void)
{
	enum msg_group_t i;
	pmsg_group group;
	for (i = 0; i < MAX_ID; i++ ) {
		group = &all_groups[i];
		memcpy(group->messages, group->defaults,
			sizeof(char*)*group->size);
	}
}

/* Free user-supplied strings when module is unloaded: */
void spk_free_user_msgs(void)
{
	enum msg_group_t n;
	int i;
	pmsg_group group;
	unsigned long flags;

	spin_lock_irqsave(&speakup_info.spinlock, flags);
	for (n = 0; n < MAX_ID; n++ ) {
		group = &all_groups[n];
		for (i = 0; i < group->size; i++) {
			if (group->messages[i] && group->messages[i] != group->defaults[i])
				kfree(group->messages[i]);
			group->messages[i] = group->defaults[i];
		}
	}
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}
