#ifndef _SPEAKUP_H
#define _SPEAKUP_H

#include "spk_types.h"
#include "i18n.h"

#define SPEAKUP_VERSION "3.1.6"
#define KEY_MAP_VER 121
#define SHIFT_TBL_SIZE 64
#define MAX_DESC_LEN 72


#define TOGGLE_0 .u.n = {NULL, 0, 0, 1, 0, 0, NULL }
#define TOGGLE_1 .u.n = {NULL, 1, 0, 1, 0, 0, NULL }
#define MAXVARLEN 15

#define SYNTH_OK 0x0001
#define B_ALPHA 0x0002
#define ALPHA 0x0003
#define B_CAP 0x0004
#define A_CAP 0x0007
#define B_NUM 0x0008
#define NUM 0x0009
#define ALPHANUM (B_ALPHA|B_NUM)
#define SOME 0x0010
#define MOST 0x0020
#define PUNC 0x0040
#define A_PUNC 0x0041
#define B_WDLM 0x0080
#define WDLM 0x0081
#define B_EXNUM 0x0100
#define CH_RPT 0x0200
#define B_CTL 0x0400
#define A_CTL (B_CTL+SYNTH_OK)
#define B_SYM 0x0800
#define B_CAPSYM (B_CAP|B_SYM)

#define IS_WDLM(x) (spk_chartab[((u_char)x)]&B_WDLM)
#define IS_CHAR(x, type) (spk_chartab[((u_char)x)]&type)
#define IS_TYPE(x, type) ((spk_chartab[((u_char)x)]&type) == type)
#define SET_DEFAULT -4
#define E_RANGE -3
#define E_TOOLONG -2
#define E_UNDEF -1

int speakup_thread(void *data);
void spk_reset_default_chars(void);
void spk_reset_default_chartab(void);
void synth_start(void);
int synth_insert_index(int value);
int synth_get_index(void);
int spk_set_key_info(const u_char *key_info);
void spk_set_punc(enum var_id_t who, const int val);
char *spk_strlwr(char *s);
char *spk_s2i(char *start, int *dest);
char *spk_s2uchar(char *in, u_char *out_val);
int speakup_kobj_init(void);
void speakup_kobj_exit(void);
u_short spk_chartab_get_value(char *keyword);
char *spk_chartab_get_name(int char_num, char *buffer);
void speakup_init_vars(int cols);
void speakup_register_var(struct var_t *var);
void speakup_unregister_var(enum var_id_t var_id);
struct st_var_header *spk_get_var_header(enum var_id_t var_id);
struct st_var_header *spk_var_header_by_name(const char *name);
int spk_set_num_var(int val, struct st_var_header *var, int how);
int spk_set_num_var_byid(enum var_id_t var_id, int val, int how);
int spk_set_string_var(const char *page, struct st_var_header *var, int len);
int spk_adjust_var(int var_id, int inc);
int spk_set_mask_bits(const char *input, const int which, const int how);
extern special_func spk_special_handler;
int spk_help_init(void);
void spk_set_vars(struct vc_data *vc);
int synth_init(char *name);
void synth_reset(void);
void synth_release(void);

void spk_do_flush(void);
void speakup_start_ttys(void);
void synth_buffer_add(char ch);
void synth_buffer_clear(void);
void speakup_init_selection( void );
void speakup_free_selection( void );
int speakup_set_selection(struct vc_data *vc, int ps, int pe, int buf_num);
int speakup_paste_selection(struct tty_struct *tty, int buffer_num);
int speakup_read_selection(int buf_num);
void speakup_register_devsynth(void);
void speakup_unregister_devsynth(void);
void synth_write(const char *buf, int count);
#define synth_write_string(buf) synth_write(buf, -1);
void spkup_write(const char *buf, int count, u_short mask);
// like synth_write, but does repeats/punctuation processing


extern wait_queue_head_t speakup_event;
extern struct kobject *speakup_kobj;
extern struct task_struct *speakup_task;
extern const u_char spk_key_defaults[];

/* Protect speakup synthesizer list */
extern struct mutex spk_mutex;
extern struct st_spk_t *speakup_console[];
extern struct spk_synth *synth;
extern char spk_pitch_buff[];
extern u_char *spk_our_keys[];
extern char spk_str_caps_start[], spk_str_caps_stop[];
extern const struct st_bits_data spk_punc_info[];
extern u_char spk_key_buf[];
extern u_char spk_close_press;
extern char *spk_characters[];
extern char *spk_default_chars[];
extern u_short spk_chartab[];
extern int spk_no_intr, spk_say_ctrl, spk_say_word_ctl, spk_punc_level;
extern int *spk_reading_punc, spk_attrib_bleep, spk_bleeps;
extern int spk_bleep_time, spk_bell_pos;
extern int spk_spell_delay, spk_cursor_time, spk_key_echo;
extern int spk_punc_mask, spk_reading_mask;
extern short spk_pitch_shift, synth_flags;
extern bool spk_quiet_boot;
extern char *synth_name;
extern int spk_delay_time, spk_trigger_time, spk_jiffy_time, spk_full_time;

/* Prototypes from fakekey.c. */
int spk_keybuff_size(void);
int speakup_add_virtual_keyboard(void);
void speakup_remove_virtual_keyboard(void);
void speakup_fake_key( unsigned int key );
bool speakup_fake_key_pressed(void);

#endif
