#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/input.h>
#include "spk_types.h"
#include "spk_priv.h"
#include "speakup.h"

static struct st_var_header var_headers[] = {
	[VERSION] = { "version", VERSION, VAR_PROC, NULL },
	[SYNTH] = { "synth_name", SYNTH, VAR_PROC, NULL },
	[KEYMAP] = { "keymap", KEYMAP, VAR_PROC, NULL },
	[SILENT] = { "silent", SILENT, VAR_PROC, NULL },
	[PUNC_SOME] = { "punc_some", PUNC_SOME, VAR_PROC, NULL },
	[PUNC_MOST] = { "punc_most", PUNC_MOST, VAR_PROC, NULL },
	[PUNC_ALL] = { "punc_all", PUNC_ALL, VAR_PROC, NULL },
	[DELIM] = { "delimiters", DELIM, VAR_PROC, NULL },
	[REPEATS] = { "repeats", REPEATS, VAR_PROC, NULL },
	[EXNUMBER] = { "ex_num", EXNUMBER, VAR_PROC, NULL },
	[CHARS] = { "characters", CHARS, VAR_PROC, NULL },
	[CHARTAB] = { "chartab", CHARTAB, VAR_PROC, NULL },
	[SYNTH_DIRECT] = { "synth_direct", SYNTH_DIRECT, VAR_PROC, NULL },
	[CAPS_START] = { "caps_start", CAPS_START, VAR_STRING, spk_str_caps_start },
	[CAPS_STOP] = { "caps_stop", CAPS_STOP, VAR_STRING, spk_str_caps_stop },
	[DELAY] = { "delay_time", DELAY, VAR_TIME, &spk_delay_time },
	[TRIGGER] = { "trigger_time", TRIGGER, VAR_TIME, &spk_trigger_time },
	[JIFFY] = { "jiffy_delta", JIFFY, VAR_TIME, &spk_jiffy_time },
	[FULL] = { "full_time", FULL, VAR_TIME, &spk_full_time },
	[SPELL_DELAY] = { "spell_delay", SPELL_DELAY, VAR_NUM, &spk_spell_delay },
	[BLEEPS] = { "bleeps", BLEEPS, VAR_NUM, &spk_bleeps },
	[ATTRIB_BLEEP] = { "attrib_bleep", ATTRIB_BLEEP, VAR_NUM, &spk_attrib_bleep },
	[BLEEP_TIME] = { "bleep_time", BLEEP_TIME, VAR_TIME, &spk_bleep_time },
	[CURSOR_TIME] = { "cursor_time", CURSOR_TIME, VAR_TIME, &spk_cursor_time },
	[PUNC_LEVEL] = { "punc_level", PUNC_LEVEL, VAR_MULTI, &spk_punc_mask },
	[READING_PUNC] = { "reading_punc", READING_PUNC, VAR_MULTI, &spk_reading_mask },
	[SAY_CONTROL] = { "say_control", SAY_CONTROL, VAR_NUM, &spk_say_ctrl },
	[SAY_WORD_CTL] = { "say_word_ctl", SAY_WORD_CTL, VAR_NUM, &spk_say_word_ctl },
	[NO_INTERRUPT] = { "no_interrupt", NO_INTERRUPT, VAR_NUM, &spk_no_intr },
	[KEY_ECHO] = { "key_echo", KEY_ECHO, VAR_NUM, &spk_key_echo },
	[BELL_POS] = { "bell_pos", BELL_POS, VAR_NUM, &spk_bell_pos },
	[RATE] = { "rate", RATE, VAR_MULTI, NULL },
	[VOICE] = { "voice", VOICE, VAR_MULTI, NULL },
	[PITCH] = { "pitch", PITCH, VAR_MULTI, NULL },
	[VOL] = { "vol", VOL, VAR_MULTI, NULL },
	[TONE] = { "tone", TONE, VAR_MULTI, NULL },
	[PUNCT] = { "punct", PUNCT, VAR_MULTI, NULL, NULL  },
	[FREQUENCY] = { "freq", FREQUENCY, VAR_MULTI, NULL },
	[LANG] = { "lang", LANG, VAR_MULTI, NULL },
	[DIRECT] = { "direct", DIRECT, VAR_NUM, NULL },
};

static struct var_t spk_vars[] = { // bell must be first
	{BELL_POS, .u.n = {NULL, 0, 0, 0, 0, 0, NULL} },
	{SPELL_DELAY, .u.n = {NULL, 0, 0, 4, 0, 0, NULL} },
	{ATTRIB_BLEEP, .u.n = {NULL, 1, 0, 3, 0, 0, NULL} },
	{BLEEPS, .u.n = {NULL, 3, 0, 3, 0, 0, NULL} },
	{BLEEP_TIME, .u.n = {NULL, 50, 10, 200, 0, 0, NULL} },
	{PUNC_LEVEL, .u.n = {NULL, 1, 0, 4, 0, 0, NULL} },
	{READING_PUNC, .u.n = {NULL, 1, 0, 4, 0, 0, NULL} },
	{CURSOR_TIME, .u.n = {NULL, 60, 25, 600, 0, 0, NULL} },
	{SAY_CONTROL, TOGGLE_0},
	{SAY_WORD_CTL, .u.n = {NULL, 0, 0, 2, 0, 0, NULL} },
	{NO_INTERRUPT, TOGGLE_0},
	{KEY_ECHO, .u.n = {NULL, 1, 0, 2, 0, 0, NULL} },
	{ PUNC_SOME, .u.p = { 1 } },
	{ PUNC_MOST, .u.p = { 2 } },
	{ PUNC_ALL, .u.p = { 3 } },
	{ DELIM, .u.p = { 4 } },
	{ REPEATS, .u.p = { 5 } },
	{ EXNUMBER, .u.p = { 6 } },
	{ DELAY, .u.n = {NULL, 100, 100, 2000, 0, 0, NULL } },
	{ TRIGGER, .u.n = {NULL, 20, 10, 200, 0, 0, NULL } },
	{ JIFFY, .u.n = {NULL, 50, 20, 200, 0, 0, NULL } },
	{ FULL, .u.n = {NULL, 400, 200, 60000, 0, 0, NULL } },
	V_LAST_VAR
};

static char *bit_symbols = "+aunsmpwxrcX";
#define MAXBIT 0x800

u_short spk_chartab_get_value(char *keyword)
{
	u_short val = 0;
	char *cp = keyword, *bp, ch;
	int i;
	while ((ch = *cp++)) {
		if (!(bp = strchr(bit_symbols, ch)))
			return 0;	
		i = bp-bit_symbols;
		val |= 1<<i;
	}
	return val;
}

char *spk_chartab_get_name(int char_num, char *buffer)
{
	char *cp = buffer, *bp = bit_symbols, ch;
	u_short mask = 1, val = spk_chartab[char_num];
	while ((ch = *bp++)) {
		if ((val&mask))
			*cp++ = ch;
		mask <<= 1;
	}
	*cp = '\0';
	return buffer;
}

void speakup_init_vars(int cols) {
	struct var_t *var = spk_vars;
	struct st_var_header *p_header;
	int i;
	spk_vars[0].u.n.high = cols;
	for (i = 0; i < MAXVARS; i++) {
		p_header = &var_headers[i];
		p_header->data = NULL;
		p_header->multi = NULL;
	}
	while (var->var_id != MAXVARS)
		speakup_register_var(var++);
}

void speakup_register_var(struct var_t *var)
{
	struct st_var_header *p_header;
	static char nothing[2] = "\0";

	BUG_ON(!var || var->var_id < 0 || var->var_id >= MAXVARS);
	p_header = &var_headers[var->var_id];
	if ((p_header->data))
		return;
	p_header->data = var;
	switch (p_header->var_type) {
	case VAR_STRING:
		spk_set_string_var(nothing, p_header, 0);
		break;
	case VAR_MULTI:
		spk_set_num_var(0, p_header, E_DEFAULT);
		spk_set_num_var(fg_console, p_header, E_MULTI);
		break;
	case VAR_NUM:
	case VAR_TIME:
		spk_set_num_var(0, p_header, E_DEFAULT);
		break;
	default:
		break;
	}
}

void speakup_unregister_var(enum var_id_t var_id)
{
	struct st_var_header *p_header;
	BUG_ON(var_id < 0 || var_id >= MAXVARS);
	p_header = &var_headers[var_id];
	if ((p_header->multi)) {
		kfree(p_header->multi);
		p_header->multi = NULL;
	}
	p_header->data = NULL;
}

struct st_var_header *spk_get_var_header(enum var_id_t var_id)
{
	struct st_var_header *p_header;
	if (var_id < 0 || var_id >= MAXVARS)
		return NULL;
	p_header = &var_headers[var_id];
	return (p_header->data ) ? p_header: NULL;
}
EXPORT_SYMBOL_GPL(spk_get_var_header);

struct st_var_header *spk_var_header_by_name(const char *name)
{
	int i;
	struct st_var_header *p_header;

	if (name == NULL) return NULL;
		for (i = 0; i < MAXVARS; i++) {
		p_header = &var_headers[i];
		if (strcmp(name, p_header->name) == 0)
			break;
	}
	if (i < MAXVARS && p_header->data)
		return p_header;
	return NULL;
}
EXPORT_SYMBOL_GPL(spk_var_header_by_name);

struct var_t *spk_get_var(enum var_id_t var_id)
{
	BUG_ON(var_id < 0 || var_id >= MAXVARS);
	return (var_headers[var_id].data);
}
EXPORT_SYMBOL_GPL(spk_get_var);

/* handlers for setting vars */

int spk_set_num_var_byid(enum var_id_t var_id, int input, int how)
{
	struct st_var_header *p_header;

	BUG_ON(var_id < 0 || var_id >= MAXVARS);
	p_header = &var_headers[var_id];
	if (p_header->var_type > VAR_TIME)
		return 0;
	return spk_set_num_var(input, p_header, how);
}
EXPORT_SYMBOL_GPL(spk_set_num_var_byid);

int spk_set_num_var(int input, struct st_var_header *var, int how)
{
	int val, ret = 0, i;
	int *p_val = (int*)var->p_val;
	int l;
	char buf[32];
	char *cp;
	struct var_t *var_data = var->data;
	if (!var_data )
		return E_UNDEF;
	switch(how) {
		case E_NEW_DEFAULT:
			var_data->u.n.default_val = input;
		// fall through
		case E_DEFAULT:
			val = var_data->u.n.default_val;
			ret = SET_DEFAULT;
			break;
		case E_SET:
			val = input;
			break;
		case E_ADJUST:
			val = var_data->u.n.value+input;
			break;
		case E_NEWCON:
			if ( !var->multi )
				return E_RANGE;
			val = var->multi[input];
			if ( val < var_data->u.n.low) {
				val = var_data->u.n.value;
				var->multi[input] = val;
			}
			if ( var_data->u.n.value == val ) return 0;
			break;
		case E_MULTI:
			if (var->var_type > VAR_MULTI)
				return 0;
			if (!var->multi)
				var->multi = kmalloc((sizeof(int)*MAX_NR_CONSOLES), GFP_KERNEL);
			if (!var->multi ) {
				pr_warn( "can't alloc multi for %s", var->name );
				return E_UNDEF;
			}
			val = var_data->u.n.value;
			for (i = 0; i < MAX_NR_CONSOLES; i++)
				var->multi[i] = -1;
			var->multi[input] = val;
			var->var_type = VAR_MULTI;
			return 0;
		case E_SINGLE:
			if ( var->var_type != VAR_MULTI)
				return 0;
			if(var->multi)
				kfree(var->multi);
			var->multi = NULL;
			var->var_type = VAR_NUM;
			return 0;
		default:
			return E_UNDEF;
	}
	if (val < var_data->u.n.low || val > var_data->u.n.high)
		return E_RANGE;
	var_data->u.n.value = val;
	if ( how < E_NEWCON &&(var->multi))
		var->multi[fg_console] = val;
	if ((p_val)) {
		if (var->var_type == VAR_TIME)
			*p_val = msecs_to_jiffies(val);
		else if (var->var_id == PUNC_LEVEL || var->var_id == READING_PUNC)
			*p_val = spk_punc_info[val].mask;
		else *p_val = val;
	}
	if (var->var_id < FIRST_SYNTH_VAR || !synth )
		return ret;
	if (!var_data->u.n.synth_fmt)
		return ret;
	if (var_data->u.n.multiplier != 0)
		val *= var_data->u.n.multiplier;
	val += var_data->u.n.offset;
	cp = (var->var_id == PITCH) ? spk_pitch_buff : buf;
	if (!var_data->u.n.out_str)
		l = sprintf(cp, var_data->u.n.synth_fmt, (int)val);
	else
		l = sprintf(cp, var_data->u.n.synth_fmt, var_data->u.n.out_str[val]);
	synth_write(cp, l);
	if (var->var_id == VOICE && (synth->new_voice))
		ret = synth->new_voice(val);
	return ret;
}
EXPORT_SYMBOL_GPL(spk_set_num_var);

int spk_set_string_var(const char *page, struct st_var_header *var, int len)
{
	int ret = 0;
	struct var_t *var_data = var->data;
	if (!var_data || !var->p_val)
		return E_UNDEF;
	if (len < 0 || len >= MAXVARLEN)
		return E_TOOLONG;
	if (!len) {
		if (!var_data->u.s.default_val)
			return 0;
		ret = SET_DEFAULT;
		strcpy((char *)var->p_val, var_data->u.s.default_val);
	} else if (page) {
		strcpy((char *)var->p_val, page);
	} else {
		pr_warn("set var oops %s\n", var->name);
	}
	return ret;
}
EXPORT_SYMBOL_GPL(spk_set_string_var);

/* set_mask_bits sets or clears the punc/delim/repeat bits,
 * if input is null uses the defaults.
 * values for how: 0 clears bits of chars supplied,
 * 1 clears all, 2 sets bits for chars */
int spk_set_mask_bits(const char *input, const int which, const int how)
{
	u_char *cp;
	short mask = spk_punc_info[which].mask;
	if (how&1) {
		for (cp = (u_char *)spk_punc_info[3].value; *cp; cp++)
			spk_chartab[*cp] &= ~mask;
	}
	cp = (u_char *)input;
	if (cp == 0)
		cp = spk_punc_info[which].value;
	else {
		for ( ; *cp; cp++) {
			if (*cp < SPACE)
				break;
			if (mask < PUNC) {
				if (!(spk_chartab[*cp]&PUNC))
					break;
			} else if (spk_chartab[*cp]&B_NUM)
				break;
		}
		if (*cp)
			return -EINVAL;
		cp = (u_char *)input;
	}
	if (how&2) {
		for ( ; *cp; cp++)
			if (*cp > SPACE)
				spk_chartab[*cp] |= mask;
	} else {
		for ( ; *cp; cp++)
			if (*cp > SPACE)
				spk_chartab[*cp] &= ~mask;
	}
	return 0;
}

char *spk_strlwr(char *s)
{
	char *p;
	if (!s )
		return NULL;

	for (p = s; *p; p++)
		*p = tolower(*p);
	return s;
}

int spk_adjust_var(int var_id, int inc)
{
	static struct st_var_header *p_header = NULL;
	int adj = E_ADJUST;
	char num_buf[32];
	char *cp = num_buf;
	char *pn;
	if (var_id >= BLEEP_TIME) {
		p_header = spk_get_var_header( var_id );
		if ( p_header == NULL || p_header->var_type > VAR_TIME )
			return -1;
		if (inc != 0 ) {
			if ( p_header->var_type == VAR_TIME )
				inc *= 5; // time in 5ms inc/dec
			if (spk_set_num_var(inc, p_header, adj) != 0)
				return -1;
		} else
			spk_close_press = 0;
	} else { // a toggle multi
		if (var_id != 0|| p_header == NULL )
			return -1;
		inc = fg_console;
		adj = (p_header->var_type == VAR_NUM) ? E_MULTI : E_SINGLE;
		if (spk_set_num_var(inc, p_header, adj) != 0)
			return -1;
	}
	if (!spk_close_press || adj != E_ADJUST) {
		for (pn = p_header->name; *pn; pn++) {
			if (*pn == '_')
				*cp++ = SPACE;
			else
				*cp++ = *pn;
		}
	}
	if (adj == E_ADJUST) {
		struct var_t *var_data = p_header->data;
		int val = var_data->u.n.value;
		if (var_data->u.n.multiplier > 0) {
			val *= var_data->u.n.multiplier;
			val += var_data->u.n.offset;
		}
		snprintf(cp, sizeof(num_buf) - (cp - num_buf), " %d ", val );
	} else if (adj == E_MULTI)
		snprintf(cp, sizeof(num_buf) - (cp - num_buf), " set to multi ");
	else // set back to single
		snprintf(cp, sizeof(num_buf) - (cp - num_buf), " set to single ");
	synth_printf("%s", num_buf);
	return 0;
}
EXPORT_SYMBOL_GPL(spk_adjust_var);

char *spk_s2uchar(char *in, u_char *out_val)
{
	int val = 0;
	while (*in && *in <= SPACE)
		in++;
	while (*in >= '0' && *in <= '9') {
		val *= 10;
		val += (*in) - '0';
		in++;
	}
	if (' ' == *in || ',' == *in)
		in++;
	if (val > 255)
		val = 0 ;
	*out_val = (u_char)val;
	return in;
}
EXPORT_SYMBOL_GPL(spk_s2uchar);

static enum var_id_t adjust_list[32];
static int list_size = 0, list_index = 0;

static int do_adjust(struct vc_data *vc, u_char type, u_char ch, u_short key)
{
	int inc = 0;
	if (key == KEY_SPACE) {
		spk_special_handler = NULL;
		spk_msg_say(MSG_DONE_VARS);
		return 1;
	}
	if (key == KEY_UP && list_index > 0)
		list_index--;
	else if (key == KEY_DOWN && list_index+1 < list_size)
		list_index++;
	else if (key == KEY_LEFT)
		inc = -1;
	else if (key == KEY_RIGHT)
		inc = 1;
	else return -1;
	spk_adjust_var(adjust_list[list_index], inc);
	return 1;
}

static enum var_id_t adjustable_vars[] = {
	ATTRIB_BLEEP, BELL_POS, BLEEPS, BLEEP_TIME, CURSOR_TIME, 
	FREQUENCY, KEY_ECHO, LANG, NO_INTERRUPT,
PITCH, PUNCT, PUNC_LEVEL, 
	RATE, READING_PUNC, SAY_CONTROL, SAY_WORD_CTL, SPELL_DELAY, 
	TONE, VOICE, VOL,MAXVARS
};

void spk_set_vars(struct vc_data *vc)
{
	struct st_var_header *p_header;
	enum var_id_t v_num, i;
		spk_msg_say(MSG_VARS_INFO);
	synth_write_string(",\n");
	spk_special_handler = do_adjust;
	if (list_size > 0) {
		spk_adjust_var(adjust_list[list_index], 0);
		return;
	}
	for (i = 0; adjustable_vars[i] < MAXVARS; i++ ) {
		v_num = adjustable_vars[i];
		p_header = &var_headers[v_num];
		if  (p_header->var_type <= VAR_TIME && p_header->data)
			adjust_list[list_size++] = v_num;
	}
	adjust_list[list_size] = 0;
	list_index = list_size/2;
	spk_adjust_var(adjust_list[list_index], 0);
}
EXPORT_SYMBOL_GPL(spk_set_vars);
